"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.outputFields = void 0;
const complex_1 = __importDefault(require("./complex"));
const object_id_1 = require("./object-id");
const schema_1 = require("./schema");
const datePatternTmp = /^\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d\.\d\d\d(Z|[+-]\d\d:?\d\d)$/.toString();
const datePattern = datePatternTmp.substring(1, datePatternTmp.length - 1);
function default_1(schema, context = 'jsonschema', options = {}) {
    const output = [];
    const base = {
        $id: options.schemaId,
        type: 'object',
        $schema: 'http://json-schema.org/draft-07/schema#',
        properties: {},
        required: [],
        additionalProperties: options.allowAdditionalFields || options.allowAdditionalFieldsNested,
    };
    outputFields(schema.fields, context, base, options);
    return base;
}
exports.default = default_1;
function outputFields(fields, context, schema, options) {
    for (const key of Object.keys(fields)) {
        const field = fields[key];
        const presentIn = field.presentIn;
        if (presentIn && !presentIn.includes(context))
            continue;
        schema.properties[key] = outputFieldFormat(field, context, options);
        const optional = options.makeEverythingOptional || isOptional(field, context);
        if (!optional)
            schema.required.push(key);
    }
}
exports.outputFields = outputFields;
function isOptional(field, context) {
    if (!field.type)
        return false;
    const fi = field;
    return fi.optional || (fi.optionalIn && fi.optionalIn.includes(context));
}
function outputFieldFormat(field, context, options) {
    if (isFullDeclaration(field)) {
        if (field.enum && field.type !== String && !(Array.isArray(field.type) && field.type[0] === String)) {
            throw new Error('Enum is only supported for strings');
        }
        const prop = asJSONSchemaProperty(field, context, options);
        if (field.enum && !Array.isArray(field.type)) {
            // tslint:disable-next-line
            ;
            prop.enum = field.enum;
        }
        if (field.jsonSchema)
            Object.assign(prop, field.jsonSchema);
        if (field.allowNull) {
            const chosenType = prop.type;
            if (typeof chosenType !== 'string')
                throw new Error('Cannot use allowNull when base type is an array');
            return {
                ...prop,
                type: [chosenType, 'null'],
                ...(field.jsonSchema || {}),
            };
        }
        return prop;
    }
    else {
        return asJSONSchemaProperty({ type: field }, context, options);
    }
}
function asJSONSchemaProperty(field, context, options) {
    const { type } = field;
    if (type === object_id_1.ObjectId)
        return { type: 'string', pattern: '^[a-fA-F0-9]{24}$' };
    if (type === String)
        return { type: 'string' };
    if (type === Date) {
        return { type: 'string', pattern: datePattern };
    }
    if (type === Boolean)
        return { type: 'boolean' };
    if (type === Number)
        return { type: 'number' };
    if (type === Object)
        return {};
    if (complex_1.default.isComplex(type)) {
        return type.outputJSONSchema(context, options, field);
    }
    if (type instanceof Array) {
        const outtype = outputFieldFormat(type[0], context, options);
        const subschema = {
            type: 'array',
            items: {
                ...outtype,
                ...(field.enum ? { enum: field.enum } : {}),
            },
        };
        return subschema;
    }
    if (schema_1.isSchema(type)) {
        return new complex_1.default(type.fields).outputJSONSchema(context, options, field);
    }
    throw new Error('Unsupported type for json-schema ' + type);
}
function isFullDeclaration(field) {
    return !!field.type;
}
function* yieldMany(items) {
    for (const item of items)
        yield item;
}
